#ifndef SAMPLE_H
#define SAMPLE_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <vector>

#include "rumd/rumd_base.h"
#include "rumd/rumd_technical.h"
#include "rumd/LogLinOutputManager.h"

#include "rumd/SimulationBox.h"
#include "rumd/ParticleData.h"
#include "rumd/NeighborList.h"
#include "rumd/KernelPlan.h"

class Integrator; class Potential;
class MoleculeData;
class EnergiesOutputManager;  class ExternalCalculator;
class ConfigurationWriterReader;

class Sample{
  friend class ConfigurationWriterReader;

private:
  unsigned int particles_per_block;
  unsigned int threads_per_particle;
  unsigned int degreesOfFreedom;
  unsigned long int blockSize;
  bool allow_tp_280_large;

  std::string trajectoryDir;
  bool managers_initialized;
  std::map<std::string, LogLinOutputManager*> outputManagers;

  LogLinOutputManager* restartManager;
  LogLinOutputManager* trajectoryManager;
  EnergiesOutputManager* energiesOutputManager; 
  std::vector<Potential*> potentialList;

  // Input/Output
  bool verbose;
  float4 CalculateCenterOfMass( float4* position, float4* velocity, int moleculeIndex );

  ConfigurationWriterReader* restartWriter;
  ConfigurationWriterReader* trajectoryWriter;
  void CreateDefaultOutputManagers();

  // the size above which default sorting method becomes XYZ
  static const unsigned int max_num_particles_sort_XY;
 protected:
  SimulationBox* simulationBox;
  bool own_sim_box;
  Integrator* itg; 
  ParticleData particleData;
  MoleculeData* moleculeData;
  KernelPlan kPlan;
  SortingScheme sortingScheme;
  
 public:
  Sample( unsigned int pb, unsigned int tp, bool allow_tp_280_large );
  Sample(const Sample& S);
  Sample* Copy();
  ~Sample();
  Sample& operator=(const Sample& S);

  void InitializeOutputManagers(unsigned long int timeStepIndex);
  void NotifyOutputManagers(unsigned long int timeStepIndex);
  void TerminateOutputManagers(); // close open files etc
  void ReadConf( const std::string& filename, bool init_itg=true );
  void ReadRestartConf(int restartBlock, unsigned int numDigitsBlkIdx=4 );
  void WriteConf(const std::string& filename, const std::string& mode="w");
  void AddOutputManager(const std::string& manager_name, LogLinOutputManager* om);
 
  float CalcF(bool calc_stresses = false, bool report_time = false);
  void IsotropicScaleSystem( float Rscal );
  void IsotropicScaleSystemCM( float Rscal );
  void AnisotropicScaleSystem( float Rscal, unsigned dir );
  void AffinelyShearSystem( float shear_strain );
  void ScaleVelocities( float factor );
  void CopySimulationDataToDevice() const;
  void CopySimulationDataFromDevice() const;
  void AddPotential(Potential* potential);
  void SubtractDOFs( unsigned i ){ degreesOfFreedom -= i; };
  void SetMoleculeData(MoleculeData* md) { moleculeData = md; }
  // Spatial sorting
  void SortParticles();
  
  // Set methods
  void SetOutputBlockSize(unsigned long int blockSize);
  void SetOutputManagerActive(const std::string &manager_name, bool active);
  void SetLogLinParameters(const std::string& manager_name, unsigned int base, unsigned int maxInterval);
  void SetOutputManagerMetaData(const std::string& manager_name, const std::string& key, bool on);
  void SetOutputManagerMetaData(const std::string& manager_name, const std::string& key, int value);

  void AddExternalCalculator(ExternalCalculator* calc);
  void RemoveExternalCalculator(ExternalCalculator* calc); 

  void SetNumberOfParticles(unsigned int num_part);
  void SetSimulationBox(SimulationBox* simBox, bool own_this_box=false);
  void SetIntegrator(Integrator* newItg);
  void SetPotential(Potential* potential);
  void SetVerbose(bool vb);
  void SetOutputDirectory(const std::string& trajectoryDir);
  void SetNumberOfDOFs( unsigned i ){ degreesOfFreedom = i; };

  void SetConfIO_Format(unsigned int ioformat);

  void SetPB_TP( unsigned int pb, unsigned int tp );
  void SetSortingScheme(SortingScheme ss);

  // Get methods.
  SimulationBox* GetSimulationBox() const { return simulationBox; }  
  ParticleData* GetParticleData() { return &particleData; }
  const ParticleData* GetParticleDataConst() const { return &particleData; }
  MoleculeData* GetMoleculeData() { return moleculeData; }
  Integrator* GetIntegrator() const { return itg; }
  std::vector<Potential*>* GetPotentials() { return &potentialList; }  
  const KernelPlan& GetKernelPlan() const { return kPlan; }
  std::string GetOutputDirectory() const { return trajectoryDir; }

  unsigned int GetParticlesPerBlock() const { return particles_per_block; }
  unsigned int GetThreadsPerParticle() const { return threads_per_particle; }
  unsigned int GetNumberOfDOFs() const { return degreesOfFreedom; }
  unsigned int GetNumberOfParticles() const { return particleData.GetNumberOfParticles(); }
  unsigned int GetNumberOfVirtualParticles() const { return particleData.GetNumberOfVirtualParticles(); }
  unsigned int GetNumberOfTypes() const { return particleData.GetNumberOfTypes(); }
  unsigned int GetNumberThisType(unsigned int type) const { return particleData.GetNumberThisType(type); }
  unsigned long int GetOutputBlockSize() const { return blockSize; }

  float GetMass(unsigned int type) const { return particleData.GetMass(type); }
  float GetMeanMass() const { float meanMass = 0; for(unsigned i=0; i < GetNumberOfTypes(); i++){ meanMass += GetMass(i); } meanMass /= GetNumberOfTypes(); return meanMass; }


  double GetPotentialEnergy(bool copy=true) const;  
  double GetVirial(bool copy=true) const;
  std::vector<double> GetStress(bool copy=true) const;
  SortingScheme GetSortingScheme() const { return sortingScheme; }
};


#endif // SAMPLE_H
