#ifndef SIMULATIONBOX_H
#define SIMULATIONBOX_H

/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/RUMD_Error.h"
#include "rumd/rumd_base.h"
#include <cmath>
#include <vector>

// The abstract Simulation Box class. Inherit this class to build your own SimulationBox.
class SimulationBox{
 
 private:
  SimulationBox(const SimulationBox&); 
  SimulationBox& operator=(const SimulationBox&); 
  
 protected:
  float* h_boxDetails;
  float* d_boxDetails;
  
 public:
  SimulationBox() : h_boxDetails(0), d_boxDetails(0){};
  virtual ~SimulationBox(){};
  
  __host__ __device__ float4 calculateDistance( float4 position1, float4 position2, float* simBoxPtr );
  __host__ __device__ float4 calculateDistanceWithImages( float4 position1, float4 position2, float4 image1, float4 image2, float* simBoxPtr );
  __host__ __device__ float4 applyBoundaryCondition( float4& position, float* simBoxPtr );
  __host__ __device__ void loadMemory( float* toMem, float* fromMem );

  virtual void ScaleBox( float factor ) = 0;
  virtual void CopyAnotherBox(const SimulationBox* boxToBeCopied) = 0;
  virtual SimulationBox* MakeCopy() = 0;
  virtual void ScaleBoxDirection( float factor, int dir ) = 0;

  // Set methods.
  virtual void SetBoxParameters( std::vector<float>& parameterList ) = 0;

  // Get methods.
  virtual float GetVolume() const { return -1.f; };
  virtual float GetLength(int dir) const { return -1.f*dir; };
  virtual float4 GetSimulationBox() const { float4 value = {0,0,0,0}; return value; };

  virtual float* GetHostPointer(){
    if(!h_boxDetails)
      throw( RUMD_Error("SimulationBox","SimulationBox","Host pointer is invalid") );
    else
      return h_boxDetails;
  }

  virtual float* GetDevicePointer(){
    if(!d_boxDetails)
      throw( RUMD_Error("SimulationBox","SimulationBox","Device pointer is invalid") );
    else
      return d_boxDetails;
  }

  virtual std::string GetInfoString(unsigned int precision) const = 0;
};

///////////////////////////////////////////////////
// Avaliable simulation boxes
////////////////////////////////////////////////////

// A rectangular simulation box that assumes periodic boundaries.
class RectangularSimulationBox : public SimulationBox{

private:
  RectangularSimulationBox(const RectangularSimulationBox&);
  RectangularSimulationBox& operator=(const RectangularSimulationBox&);
  
  void AllocateBoxDetails();

protected:
  void CopyBoxDetailsToGPU();

 public:
  RectangularSimulationBox();
  RectangularSimulationBox( float Lx, float Ly, float Lz );
  ~RectangularSimulationBox();   

  void CopyAnotherBox(const SimulationBox* box_to_copy);
  SimulationBox* MakeCopy();
  
  __host__ __device__ float4 calculateDistance( float4 position1, float4 position2, float* simBoxPtr ){
    float dr_x = position1.x - position2.x;
    float dr_y = position1.y - position2.y;
    float dr_z = position1.z - position2.z;

    if(periodicInX)
      dr_x -= simBoxPtr[1] * rintf( dr_x * simBoxPtr[4] );
    if(periodicInY)
      dr_y -= simBoxPtr[2] * rintf( dr_y * simBoxPtr[5] );
    if(periodicInZ)
      dr_z -= simBoxPtr[3] * rintf( dr_z * simBoxPtr[6] );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }

  __host__ __device__ float4 calculateDistanceMoved( float4 position, float4 last_position, float* simBoxPtr, float2*  ){
    return calculateDistance(position, last_position, simBoxPtr);
  }

  __host__ __device__ float4 calculateDistanceWithImages( float4 position1, float4 position2, float4 image1, float4 image2, float* simBoxPtr ){
    float dr_x = (position1.x - position2.x) + simBoxPtr[1] * ( image1.x - image2.x );
    float dr_y = (position1.y - position2.y) + simBoxPtr[2] * ( image1.y - image2.y );
    float dr_z = (position1.z - position2.z) + simBoxPtr[3] * ( image1.z - image2.z );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }
  
  __host__ __device__ float4 applyBoundaryCondition( float4& position, float* simBoxPtr ){
    float4 image = { rintf( position.x * simBoxPtr[4] ), 
		     rintf( position.y * simBoxPtr[5] ), 
		     rintf( position.z * simBoxPtr[6] ), 
		     0.f };
    
    // Periodic boundary condition.
    if(periodicInX)
      position.x -= simBoxPtr[1] * image.x;
    if(periodicInY)
      position.y -= simBoxPtr[2] * image.y;
    if(periodicInZ)
      position.z -= simBoxPtr[3] * image.z;
    
    return image; 
  }
  
  __host__ __device__ void loadMemory( float* toMem, float* fromMem ){
    if((unsigned) fromMem[0]){
      toMem[1] = fromMem[1]; // boxLength
      toMem[2] = toMem[1]; 
      toMem[3] = toMem[1];
      toMem[4] = 1.f / toMem[1]; // 1 / boxLength
      toMem[5] = toMem[4];
      toMem[6] = toMem[4];
    }
    else{
      for(unsigned int i=0; i < simulationBoxSize; i++)
	toMem[i] = fromMem[i];
    }
  }
  
  void ScaleBox( float factor );
  void ScaleBoxDirection( float factor, int dir );
  
  // Set methods
  void SetBoxDetails( float* details );
  void SetBoxParameters( std::vector<float>& parameterList );

  // Get methods
  float GetVolume() const { return h_boxDetails[1] * h_boxDetails[2] * h_boxDetails[3]; }
  float GetLength(int dir) const { return h_boxDetails[dir+1]; }

  float4 GetSimulationBox() const { float4 value = { h_boxDetails[1], h_boxDetails[2], h_boxDetails[3], GetVolume() };  return value; }

  std::string GetInfoString(unsigned int precision) const;
};


////////////////////////////////////////////////////////////////
// A rectangular simulation box that assumes periodic boundaries but with 
// Lees-Edwards boundary conditions to allow relative shear of the box and
// its images.
////////////////////////////////////////////////////////////////

class LeesEdwardsSimulationBox : public RectangularSimulationBox{

 private:
  LeesEdwardsSimulationBox(const LeesEdwardsSimulationBox&);
  LeesEdwardsSimulationBox& operator=(const LeesEdwardsSimulationBox&);
  
  double box_shift_dbl;
  
 public:
  LeesEdwardsSimulationBox();
  LeesEdwardsSimulationBox( float Lx, float Ly, float Lz, float boxShift );
  LeesEdwardsSimulationBox(const SimulationBox* box_to_copy);
  ~LeesEdwardsSimulationBox() {}

  void SetBoxParameters( std::vector<float>& parameterList );

  SimulationBox* MakeCopy();
  void CopyAnotherBox(const SimulationBox* box_to_copy);

  __host__ __device__ float4 calculateDistance( float4 position1, float4 position2, float* simBoxPtr ){
    float dr_x = position1.x - position2.x;
    float dr_y = position1.y - position2.y;
    float dr_z = position1.z - position2.z;

    float y_wrap = rintf( dr_y * simBoxPtr[5] );

    dr_y -= simBoxPtr[2] * y_wrap;
    dr_x -= y_wrap * simBoxPtr[7];

    dr_x -= simBoxPtr[1] * rintf( dr_x * simBoxPtr[4] );
    dr_z -= simBoxPtr[3] * rintf( dr_z * simBoxPtr[6] );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }

  __host__ __device__ float4 calculateDistanceMoved( float4 position, float4 last_position, float* simBoxPtr, float2* dStrain ){
    // this function calculates the (square of the) non-affine displacment
    float deltaStrain = dStrain[0].x;
    float dr_x = position.x - last_position.x;
    float dr_y = position.y - last_position.y;
    float dr_z = position.z - last_position.z;

    // some of these will be reused
    float Lx = simBoxPtr[1];
    float Ly = simBoxPtr[2];
    float boxShift = simBoxPtr[7];
    float Lx_inv = simBoxPtr[4];
    float Ly_inv = simBoxPtr[5];

    float y_wrap = rintf( dr_y * Ly_inv );


    dr_y -= Ly * y_wrap;
    dr_z -= simBoxPtr[3] * rintf( dr_z * simBoxPtr[6] );
    // x-component : subtract off the affine part
    dr_x -= ( y_wrap * boxShift + (y_wrap*Ly+position.y) * deltaStrain );
    dr_x -= Lx * rintf( dr_x * Lx_inv);

    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }
  
  __host__ __device__ float4 calculateDistanceWithImages( float4 position1, float4 position2, float4 image1, float4 image2, float* simBoxPtr ){
    
    float dr_x = (position1.x - position2.x) + simBoxPtr[1] * ( image1.x - image2.x ) + simBoxPtr[7] * ( image1.y - image2.y ) ;
    
    float dr_y = (position1.y - position2.y) + simBoxPtr[2] * ( image1.y - image2.y );
    float dr_z = (position1.z - position2.z) + simBoxPtr[3] * ( image1.z - image2.z );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }
  
  __host__ __device__ float4 applyBoundaryCondition( float4& position, float* simBoxPtr ){
    float y_wrap = rintf( position.y * simBoxPtr[5] );
    position.x -= simBoxPtr[7] * y_wrap;

    float4 image = { rintf( position.x * simBoxPtr[4] ), 
		     y_wrap,
		     rintf( position.z * simBoxPtr[6] ), 
		     0.f };
    
    // Periodic boundary condition.
    position.x -= simBoxPtr[1] * image.x;
    position.y -= simBoxPtr[2] * image.y;
    position.z -= simBoxPtr[3] * image.z;
    
    return image; 
  }
  
  __host__ __device__ void loadMemory( float* toMem, float* fromMem ){
    if((unsigned) fromMem[0]){
      toMem[1] = fromMem[1]; // boxLength
      toMem[2] = toMem[1]; 
      toMem[3] = toMem[1];
      toMem[4] = 1.f / toMem[1]; // 1 / boxLength
      toMem[5] = toMem[4];
      toMem[6] = toMem[4];
      toMem[7] = fromMem[7];
    }
    else{
      for(unsigned int i=0; i < simulationBoxSize; i++)
	toMem[i] = fromMem[i];
    }
  }
  
  void ScaleBox( float factor );

  void IncrementBoxStrain(float inc_boxStrain);
  
  // Set methods
  void SetBoxShift(double set_boxShift);
  void ScaleBoxDirection( float factor, int dir );


  // Get methods
  double GetBoxShift() const { return box_shift_dbl; }
  float GetLength(int dir) const { return h_boxDetails[dir+1]; }

  std::string GetInfoString(unsigned int precision) const;
};

class SimulationBoxFactory{
 public:
  SimulationBoxFactory(){}
  virtual ~SimulationBoxFactory() {};
  SimulationBox* CreateSimulationBox(const std::string& infoStr);
 private:
};

#endif // SIMULATIONBOX_H
